use crate::app::{App, ShowEverything};
use crate::common::CommonState;
use crate::game::{DrawBaselayer, State, Transition, WizardState};
use crate::render::DrawOptions;
use ezgui::{
    hotkey, lctrl, Btn, Choice, Color, Composite, Drawable, EventCtx, GeomBatch, GfxCtx,
    HorizontalAlignment, Key, Line, Outcome, RewriteColor, Text, VerticalAlignment, Widget,
};
use geom::{Distance, LonLat, PolyLine, Polygon, Pt2D, Ring};
use serde::{Deserialize, Serialize};
use sim::DontDrawAgents;

// TODO This is a really great example of things that ezgui ought to make easier. Maybe a radio
// button-ish thing to start?

// Good inspiration: http://sfo-assess.dha.io/, https://github.com/mapbox/storytelling,
// https://storymap.knightlab.com/

pub struct StoryMapEditor {
    composite: Composite,
    story: StoryMap,
    mode: Mode,
    dirty: bool,

    // Index into story.markers
    // TODO Stick in Mode::View?
    hovering: Option<usize>,
}

enum Mode {
    View,
    PlacingMarker,
    Dragging(Pt2D, usize),
    Editing(usize, Composite),
    Freehand(Option<Lasso>),
}

impl StoryMapEditor {
    pub fn new(ctx: &mut EventCtx, app: &App) -> Box<dyn State> {
        let story = StoryMap::new();
        let mode = Mode::View;
        let dirty = false;
        Box::new(StoryMapEditor {
            composite: make_panel(ctx, app, &story, &mode, dirty),
            story,
            mode,
            dirty,
            hovering: None,
        })
    }

    fn redo_panel(&mut self, ctx: &mut EventCtx, app: &App) {
        self.composite = make_panel(ctx, app, &self.story, &self.mode, self.dirty);
    }
}

impl State for StoryMapEditor {
    fn event(&mut self, ctx: &mut EventCtx, app: &mut App) -> Transition {
        match self.mode {
            Mode::View => {
                ctx.canvas_movement();

                if ctx.redo_mouseover() {
                    self.hovering = None;
                    if let Some(pt) = ctx.canvas.get_cursor_in_map_space() {
                        self.hovering = self
                            .story
                            .markers
                            .iter()
                            .position(|m| m.hitbox.contains_pt(pt));
                    }
                }
                if let Some(idx) = self.hovering {
                    if ctx
                        .input
                        .key_pressed(Key::LeftControl, "hold to move this marker")
                    {
                        self.mode =
                            Mode::Dragging(ctx.canvas.get_cursor_in_map_space().unwrap(), idx);
                    } else if app.per_obj.left_click(ctx, "edit marker") {
                        self.mode =
                            Mode::Editing(idx, self.story.markers[idx].make_editor(ctx, app));
                    }
                }
            }
            Mode::PlacingMarker => {
                ctx.canvas_movement();

                if let Some(pt) = ctx.canvas.get_cursor_in_map_space() {
                    if app.primary.map.get_boundary_polygon().contains_pt(pt)
                        && app.per_obj.left_click(ctx, "place a marker here")
                    {
                        let idx = self.story.markers.len();
                        self.story
                            .markers
                            .push(Marker::new(ctx, vec![pt], String::new()));
                        self.dirty = true;
                        self.redo_panel(ctx, app);
                        self.mode =
                            Mode::Editing(idx, self.story.markers[idx].make_editor(ctx, app));
                    }
                }
            }
            Mode::Dragging(ref mut last_pt, idx) => {
                if ctx.redo_mouseover() {
                    if let Some(pt) = ctx.canvas.get_cursor_in_map_space() {
                        if app.primary.map.get_boundary_polygon().contains_pt(pt) {
                            let dx = pt.x() - last_pt.x();
                            let dy = pt.y() - last_pt.y();
                            *last_pt = pt;
                            self.story.markers[idx] = Marker::new(
                                ctx,
                                self.story.markers[idx]
                                    .pts
                                    .iter()
                                    .map(|pt| pt.offset(dx, dy))
                                    .collect(),
                                self.story.markers[idx].event.clone(),
                            );
                            self.dirty = true;
                            self.redo_panel(ctx, app);
                        }
                    }
                }

                if ctx.input.key_released(Key::LeftControl) {
                    self.mode = Mode::View;
                }
            }
            Mode::Editing(idx, ref mut composite) => {
                ctx.canvas_movement();
                match composite.event(ctx) {
                    Some(Outcome::Clicked(x)) => match x.as_ref() {
                        "close" => {
                            self.mode = Mode::View;
                            self.redo_panel(ctx, app);
                        }
                        "confirm" => {
                            self.story.markers[idx] = Marker::new(
                                ctx,
                                self.story.markers[idx].pts.clone(),
                                composite.text_box("event"),
                            );
                            self.dirty = true;
                            self.mode = Mode::View;
                            self.redo_panel(ctx, app);
                        }
                        "delete" => {
                            self.mode = Mode::View;
                            self.hovering = None;
                            self.story.markers.remove(idx);
                            self.dirty = true;
                            self.redo_panel(ctx, app);
                        }
                        _ => unreachable!(),
                    },
                    None => {}
                }
            }
            Mode::Freehand(None) => {
                if let Some(pt) = ctx.canvas.get_cursor_in_map_space() {
                    if ctx.input.left_mouse_button_pressed() {
                        self.mode = Mode::Freehand(Some(Lasso::new(pt)));
                    }
                }
            }
            Mode::Freehand(Some(ref mut lasso)) => {
                if let Some(result) = lasso.event(ctx) {
                    let idx = self.story.markers.len();
                    self.story
                        .markers
                        .push(Marker::new(ctx, result.into_points(), String::new()));
                    self.dirty = true;
                    self.redo_panel(ctx, app);
                    self.mode = Mode::Editing(idx, self.story.markers[idx].make_editor(ctx, app));
                }
            }
        }

        match self.composite.event(ctx) {
            Some(Outcome::Clicked(x)) => match x.as_ref() {
                "close" => {
                    // TODO autosave
                    return Transition::Pop;
                }
                "save" => {
                    if self.story.name == "new story" {
                        return Transition::Push(WizardState::new(Box::new(|wiz, ctx, _| {
                            let name = wiz.wrap(ctx).input_string("Name this story map")?;
                            Some(Transition::PopWithData(Box::new(move |state, ctx, app| {
                                let editor = state.downcast_mut::<StoryMapEditor>().unwrap();
                                editor.story.name = name;
                                editor.story.save(app);
                                editor.dirty = false;
                                editor.redo_panel(ctx, app);
                            })))
                        })));
                    } else {
                        self.story.save(app);
                        self.dirty = false;
                        self.redo_panel(ctx, app);
                    }
                }
                "load" => {
                    // TODO autosave
                    let current = self.story.name.clone();
                    let btn = self.composite.rect_of("load").clone();
                    return Transition::Push(WizardState::new(Box::new(move |wiz, ctx, app| {
                        let (_, raw) = wiz.wrap(ctx).choose_exact(
                            (
                                HorizontalAlignment::Centered(btn.center().x),
                                VerticalAlignment::Below(btn.y2 + 15.0),
                            ),
                            None,
                            || {
                                let mut list = Vec::new();
                                for (name, story) in abstutil::load_all_objects::<RecordedStoryMap>(
                                    "../data/player/stories".to_string(),
                                ) {
                                    if story.name == current {
                                        continue;
                                    }
                                    // TODO Argh, we can't make StoryMap cloneable, so redo some
                                    // work
                                    let gps_bounds = app.primary.map.get_gps_bounds();
                                    if story
                                        .markers
                                        .iter()
                                        .all(|(pts, _)| gps_bounds.try_convert(pts).is_some())
                                    {
                                        list.push(Choice::new(name, story));
                                    }
                                }
                                list.push(Choice::new(
                                    "new story",
                                    RecordedStoryMap {
                                        name: "new story".to_string(),
                                        markers: Vec::new(),
                                    },
                                ));
                                list
                            },
                        )?;
                        let story = StoryMap::load(ctx, app, raw).unwrap();
                        Some(Transition::PopWithData(Box::new(move |state, ctx, app| {
                            let editor = state.downcast_mut::<StoryMapEditor>().unwrap();
                            editor.story = story;
                            editor.dirty = false;
                            editor.redo_panel(ctx, app);
                        })))
                    })));
                }
                "new marker" => {
                    self.hovering = None;
                    self.mode = Mode::PlacingMarker;
                    self.redo_panel(ctx, app);
                }
                "draw freehand" => {
                    self.hovering = None;
                    self.mode = Mode::Freehand(None);
                    self.redo_panel(ctx, app);
                }
                "pan" => {
                    self.mode = Mode::View;
                    self.redo_panel(ctx, app);
                }
                _ => unreachable!(),
            },
            None => {}
        }

        Transition::Keep
    }

    fn draw_baselayer(&self) -> DrawBaselayer {
        DrawBaselayer::Custom
    }

    fn draw(&self, g: &mut GfxCtx, app: &App) {
        let mut opts = DrawOptions::new();
        opts.label_buildings = true;
        app.draw(g, opts, &DontDrawAgents {}, &ShowEverything::new());

        match self.mode {
            Mode::PlacingMarker => {
                if g.canvas.get_cursor_in_map_space().is_some() {
                    let batch = GeomBatch::screenspace_svg(
                        g.prerender,
                        "../data/system/assets/timeline/goal_pos.svg",
                    )
                    .centered_on(g.canvas.get_cursor().to_pt())
                    .color(RewriteColor::Change(Color::hex("#5B5B5B"), Color::GREEN));
                    g.fork_screenspace();
                    batch.draw(g);
                    g.unfork();
                }
            }
            Mode::Editing(_, ref composite) => {
                composite.draw(g);
            }
            Mode::Freehand(Some(ref lasso)) => {
                lasso.draw(g);
            }
            _ => {}
        }

        for (idx, m) in self.story.markers.iter().enumerate() {
            if self.hovering == Some(idx) {
                m.draw_hovered(g, app);
            } else {
                g.redraw(&m.draw);
            }
        }

        self.composite.draw(g);
        CommonState::draw_osd(g, app);
    }
}

fn make_panel(
    ctx: &mut EventCtx,
    app: &App,
    story: &StoryMap,
    mode: &Mode,
    dirty: bool,
) -> Composite {
    Composite::new(
        Widget::col(vec![
            Widget::row(vec![
                Line("Story map editor")
                    .small_heading()
                    .draw(ctx)
                    .margin_right(5),
                Widget::draw_batch(
                    ctx,
                    GeomBatch::from(vec![(Color::WHITE, Polygon::rectangle(2.0, 30.0))]),
                )
                .margin_right(5),
                Btn::text_fg(format!("{} ↓", story.name))
                    .build(ctx, "load", lctrl(Key::L))
                    .margin_right(5),
                if dirty {
                    Btn::svg_def("../data/system/assets/tools/save.svg").build(
                        ctx,
                        "save",
                        lctrl(Key::S),
                    )
                } else {
                    Widget::draw_svg_transform(
                        ctx,
                        "../data/system/assets/tools/save.svg",
                        RewriteColor::ChangeAlpha(0.5),
                    )
                }
                .margin_right(5),
                Btn::plaintext("X")
                    .build(ctx, "close", hotkey(Key::Escape))
                    .align_right(),
            ]),
            Widget::row(vec![
                if let Mode::PlacingMarker = mode {
                    Widget::draw_svg_transform(
                        ctx,
                        "../data/system/assets/timeline/goal_pos.svg",
                        RewriteColor::Change(Color::hex("#5B5B5B"), Color::hex("#4CA7E9")),
                    )
                } else {
                    Btn::svg_def("../data/system/assets/timeline/goal_pos.svg").build(
                        ctx,
                        "new marker",
                        hotkey(Key::M),
                    )
                },
                if let Mode::View = mode {
                    Widget::draw_svg_transform(
                        ctx,
                        "../data/system/assets/tools/pan.svg",
                        RewriteColor::ChangeAll(Color::hex("#4CA7E9")),
                    )
                } else {
                    Btn::svg_def("../data/system/assets/tools/pan.svg").build(
                        ctx,
                        "pan",
                        hotkey(Key::Escape),
                    )
                },
                match mode {
                    Mode::Freehand(_) => Widget::draw_svg_transform(
                        ctx,
                        "../data/system/assets/tools/select.svg",
                        RewriteColor::ChangeAll(Color::hex("#4CA7E9")),
                    ),
                    _ => Btn::svg_def("../data/system/assets/tools/select.svg").build(
                        ctx,
                        "draw freehand",
                        hotkey(Key::P),
                    ),
                },
            ])
            .evenly_spaced(),
        ])
        .padding(16)
        .bg(app.cs.panel_bg),
    )
    .aligned(HorizontalAlignment::Center, VerticalAlignment::Top)
    .build(ctx)
}

#[derive(Clone, Serialize, Deserialize)]
struct RecordedStoryMap {
    name: String,
    markers: Vec<(Vec<LonLat>, String)>,
}
impl abstutil::Cloneable for RecordedStoryMap {}

struct StoryMap {
    name: String,
    markers: Vec<Marker>,
}

struct Marker {
    pts: Vec<Pt2D>,
    event: String,
    hitbox: Polygon,
    draw: Drawable,
}

impl StoryMap {
    fn new() -> StoryMap {
        StoryMap {
            name: "new story".to_string(),
            markers: Vec::new(),
        }
    }

    fn load(ctx: &mut EventCtx, app: &App, story: RecordedStoryMap) -> Option<StoryMap> {
        let mut markers = Vec::new();
        for (gps_pts, event) in story.markers {
            let pts = app.primary.map.get_gps_bounds().try_convert(&gps_pts)?;
            markers.push(Marker::new(ctx, pts, event));
        }
        Some(StoryMap {
            name: story.name,
            markers,
        })
    }

    fn save(&self, app: &App) {
        let story = RecordedStoryMap {
            name: self.name.clone(),
            markers: self
                .markers
                .iter()
                .map(|m| {
                    (
                        app.primary.map.get_gps_bounds().must_convert_back(&m.pts),
                        m.event.clone(),
                    )
                })
                .collect(),
        };
        abstutil::write_json(
            format!("../data/player/stories/{}.json", story.name),
            &story,
        );
    }
}

impl Marker {
    fn new(ctx: &mut EventCtx, pts: Vec<Pt2D>, event: String) -> Marker {
        let mut batch = GeomBatch::new();

        let hitbox = if pts.len() == 1 {
            batch.append(
                GeomBatch::screenspace_svg(
                    ctx.prerender,
                    "../data/system/assets/timeline/goal_pos.svg",
                )
                .scale(2.0)
                .centered_on(pts[0])
                .color(RewriteColor::Change(
                    Color::hex("#5B5B5B"),
                    Color::hex("#FE3D00"),
                )),
            );
            batch.append(
                Text::from(Line(&event))
                    .with_bg()
                    .render_to_batch(ctx.prerender)
                    .scale(0.5)
                    .centered_on(pts[0]),
            );
            batch.unioned_polygon()
        } else {
            let poly = Polygon::new(&pts);
            batch.push(Color::RED.alpha(0.8), poly.clone());
            if let Some(o) = poly.maybe_to_outline(Distance::meters(1.0)) {
                batch.push(Color::RED, o);
            }
            // TODO Refactor
            batch.append(
                Text::from(Line(&event))
                    .with_bg()
                    .render_to_batch(ctx.prerender)
                    .scale(0.5)
                    .centered_on(poly.polylabel()),
            );
            poly
        };
        Marker {
            pts,
            event,
            hitbox,
            draw: ctx.upload(batch),
        }
    }

    fn draw_hovered(&self, g: &mut GfxCtx, app: &App) {
        let mut batch = GeomBatch::new();
        if self.pts.len() == 1 {
            batch.append(
                GeomBatch::screenspace_svg(
                    g.prerender,
                    "../data/system/assets/timeline/goal_pos.svg",
                )
                .scale(2.0)
                .centered_on(self.pts[0])
                .color(RewriteColor::Change(Color::hex("#5B5B5B"), app.cs.hovering)),
            );
            batch.append(
                Text::from(Line(&self.event))
                    .with_bg()
                    .render_to_batch(g.prerender)
                    .scale(0.75)
                    .centered_on(self.pts[0]),
            );
        } else {
            batch.push(app.cs.hovering, Polygon::new(&self.pts));
            // TODO Refactor plz
            batch.append(
                Text::from(Line(&self.event))
                    .with_bg()
                    .render_to_batch(g.prerender)
                    .scale(0.75)
                    .centered_on(self.hitbox.polylabel()),
            );
        }
        batch.draw(g);
    }

    fn make_editor(&self, ctx: &mut EventCtx, app: &App) -> Composite {
        Composite::new(
            Widget::col(vec![
                Widget::row(vec![
                    Line("Editing marker").small_heading().draw(ctx),
                    Btn::plaintext("X")
                        .build(ctx, "close", hotkey(Key::Escape))
                        .align_right(),
                ]),
                Btn::text_fg("delete").build_def(ctx, None),
                Widget::text_entry(ctx, self.event.clone(), true).named("event"),
                Btn::text_fg("confirm").build_def(ctx, hotkey(Key::Enter)),
            ])
            .padding(16)
            .bg(app.cs.panel_bg),
        )
        .build(ctx)
    }
}

// TODO This should totally be an ezgui tool
// TODO Simplify points
struct Lasso {
    pl: PolyLine,
}

impl Lasso {
    fn new(pt: Pt2D) -> Lasso {
        Lasso {
            pl: PolyLine::new(vec![pt, pt.offset(0.1, 0.0)]),
        }
    }

    fn event(&mut self, ctx: &mut EventCtx) -> Option<Ring> {
        if ctx.input.left_mouse_button_released() {
            return Some(simplify(self.pl.points().clone()));
        }
        if ctx.redo_mouseover() {
            if let Some(pt) = ctx.canvas.get_cursor_in_map_space() {
                if pt != self.pl.last_pt() {
                    // Did we make a crossing?
                    if let Some((hit, _)) = self
                        .pl
                        .intersection(&PolyLine::new(vec![self.pl.last_pt(), pt]))
                    {
                        if let Some(slice) = self.pl.get_slice_starting_at(hit) {
                            return Some(simplify(slice.into_points()));
                        }
                    }

                    let mut pts = self.pl.points().clone();
                    pts.push(pt);
                    self.pl = PolyLine::new(pts);
                }
            }
        }
        None
    }

    fn draw(&self, g: &mut GfxCtx) {
        g.draw_polygon(
            Color::RED.alpha(0.8),
            &self
                .pl
                .make_polygons(Distance::meters(5.0) / g.canvas.cam_zoom),
        );
    }
}

fn simplify(mut raw: Vec<Pt2D>) -> Ring {
    // TODO This is eating some of the shapes entirely. Wasn't meant for this.
    if false {
        let pts = raw
            .into_iter()
            .map(|pt| lttb::DataPoint::new(pt.x(), pt.y()))
            .collect();
        let mut downsampled = Vec::new();
        for pt in lttb::lttb(pts, 50) {
            downsampled.push(Pt2D::new(pt.x, pt.y));
        }
        downsampled.push(downsampled[0]);
        Ring::new(downsampled)
    } else {
        raw.push(raw[0]);
        Ring::new(raw)
    }
}
